package main

import (
	"crypto/md5"
	"encoding/hex"
	"fmt"
	"image"
	_ "image/gif"
	_ "image/jpeg"
	_ "image/png"
	"io/ioutil"
	"net/http"
	"net/url"
	"os"
	"sort"
	"strings"
	"time"
	"unicode/utf8"
	"github.com/nfnt/resize"
)

const APP_ID string = "14df7284a4af4c72bb2dc0ca8b820a33"
const APP_KEY string = "d0959fea394e437786d8db0ec94977b4"

const DOTS_PER_LINE          int = 384	// 一行的打印宽度(点)。58mm打印机为384，80mm打印机为576
const MAX_COLUMNS            int = 6	// 分列打印最大列数
         
const ALIGN_LEFT             int = 0	// 左对齐
const ALIGN_CENTER           int = 1	// 居中对齐
const ALIGN_RIGHT            int = 2	// 右对齐
         
const HRI_POS_ABOVE          int = 1	// 条码文本在条码上方
const HRI_POS_BELOW          int = 2	// 条码文本在条码下方

const COLUMN_FLAG_BW_REVERSE int = 1	// 分列打印标志：黑白反转

var orderContent string
var widthOfColumns [MAX_COLUMNS]int
var flagOfColumns [MAX_COLUMNS]int
var charHSize int = 1

func HexStrToBytes(str string) []byte {
	bytes, _ := hex.DecodeString(str)
	return bytes
}

/*
 * 生成签名。
 */
func GenerateSign(formData url.Values) string {
	keys := []string{}
	for k, _ := range formData {
		keys = append(keys, k)
	}
	sort.Strings(keys)

	s := ""
	for _, k := range keys {
		s += k + "=" + formData[k][0] + "&"
	}
	s = s[:(len(s) - 1)]
	s += APP_KEY

	h := md5.New()
	h.Write([]byte(s))
	return strings.ToUpper(hex.EncodeToString(h.Sum(nil)))
}

/*
 * 使用直推模式打印小票内容。
 */
func PushToPrinter(msn string) {
	timestamp := time.Now().Unix()
	voiceCnt := 1
	voice := "您有新的订单"
	orderCnt := 1
	orderType := 1

	formData := make(url.Values)
	formData["app_id"]    = []string{APP_ID}
	formData["msn"]       = []string{msn}
	formData["timestamp"] = []string{fmt.Sprintf("%d", timestamp)}
	formData["pushId"]    = []string{fmt.Sprintf("%s_%010d", msn, timestamp)}
	formData["voiceCnt"]  = []string{fmt.Sprintf("%d", voiceCnt)}
	formData["voice"]     = []string{voice}
	formData["orderCnt"]  = []string{fmt.Sprintf("%d", orderCnt)}
	formData["orderType"] = []string{fmt.Sprintf("%d", orderType)}
	formData["orderData"] = []string{orderContent}
	formData["sign"]      = []string{GenerateSign(formData)}
	orderContent = ""

	req, err := http.PostForm("https://openapi.sunmi.com/v1/printer/pushContent", formData)
	if err != nil {
		fmt.Println(err.Error())
		return
	}
	defer req.Body.Close()
	body, err := ioutil.ReadAll(req.Body)
	if err != nil {
		fmt.Println(err.Error())
		return
	}
	fmt.Println(string(body))
}

//////////////////////////////////////////////////
// 基本ESC/POS指令
//////////////////////////////////////////////////

/*
 * 在小票内容中追加文本字符串。
 */
func AppendText(str string) {
	if utf8.RuneCountInString(str) > 0 {
		orderContent += hex.EncodeToString([]byte(str))
	}
}

/*
 * [LF] 打印缓冲区里的内容并走纸 n 行。
 */
func LineFeed(n int) {
	for i := 0; i < n; i++ {
		orderContent += "0a"
	}
}

/*
 * [ESC @] 恢复默认设置 (行间距、字符效果等)。
 */
func RestoreDefaultSettings() {
	charHSize = 1
	orderContent += "1b40"
}

/*
 * [ESC 2] 恢复默认的行间距。
 */
func RestoreDefaultLineSpacing() {
	orderContent += "1b32"
}

/*
 * [ESC 3] 设置行间距。
 */
func SetLineSpacing(n int) {
	if n >= 0 && n <= 255 {
		orderContent += fmt.Sprintf("1b33%02x", n)
	}
}

/*
 * [ESC !] 选择加粗、倍高、倍宽模式。
 */
func SetPrintModes(bold, double_h, double_w bool) {
	n := 0
	if bold {
		n |= 8
	}
	if double_h {
		n |= 16
	}
	if double_w {
		n |= 32
		charHSize = 2
	} else {
		charHSize = 1
	}
	orderContent += fmt.Sprintf("1b21%02x", n)
}

/*
 * [GS !] 选择字符高度和宽度倍数。
 */
func SetCharacterSize(h, w int) {
	n := 0
	if h >= 1 && h <= 8 {
		n |= (h - 1)
	}
	if w >= 1 && w <= 8 {
		n |= (w - 1) << 4
		charHSize = w
	}
	orderContent += fmt.Sprintf("1d21%02x", n)
}

/*
 * [HT] 跳到下 n 个制表符位置。
 */
func HorizontalTab(n int) {
	for i := 0; i < n; i++ {
		orderContent += "09"
	}
}

/*
 * [ESC $] 跳到绝对位置。
 */
func SetAbsolutePrintPosition(n int) {
	if n >= 0 && n <= 65535 {
		orderContent += fmt.Sprintf("1b24%02x%02x", (n & 0xFF), ((n >> 8) & 0xFF))
	}
}

/*
 * [ESC \] 跳到相对位置。
 */
func SetRelativePrintPosition(n int) {
	if n >= -32768 && n <= 32767 {
		orderContent += fmt.Sprintf("1b5c%02x%02x", (n & 0xFF), ((n >> 8) & 0xFF))
	}
}

/*
 * [ESC a] 设置对齐方式。
 */
func SetAlignment(n int) {
	if n >= 0 && n <= 2 {
		orderContent += fmt.Sprintf("1b61%02x", n)
	}
}

/*
 * [GS B] 设置黑白反转模式。
 */
func SetBlackWhiteReverseMode(enabled bool) {
	var m int
	if enabled {
		m = 1
	} else {
		m = 0
	}
	orderContent += fmt.Sprintf("1d42%02x", m)
}

/*
 * [ESC {] 设置上下颠倒模式。
 */
func SetUpsideDownMode(enabled bool) {
	var m int
	if enabled {
		m = 1
	} else {
		m = 0
	}
	orderContent += fmt.Sprintf("1b7b%02x", m)
}

/*
 * [GS V m] 切纸。
 */
func CutPaper(fullCut bool) {
	var m int
	if fullCut {
		m = 48
	} else {
		m = 49
	}
	orderContent += fmt.Sprintf("1d56%02x", m)
}

/*
 * [GS V m n] 延迟切纸。
 * 发送该指令后，打印机不会马上切纸，而是等到后续走纸或打印了 (d + n) 点的距离后再切纸。
 * 其中 d 为切刀与打印头之间的距离，n 为 0~255。
 */
func PostponedCutPaper(fullCut bool, n int) {
	var m int
	if fullCut {
		m = 97
	} else {
		m = 98
	}
	if n >= 0 && n <= 255 {
		orderContent += fmt.Sprintf("1d56%02x%02x", m, n)
	}
}

//////////////////////////////////////////////////
// 分列打印
//////////////////////////////////////////////////

/*
 * 返回一个字符的宽度。0表示未知。
 */
func WidthOfChar(c rune) int {
	switch {
	case (c >= 0x00020 && c <= 0x0036F) ||
		 (c >= 0x0FF61 && c <= 0x0FF9F):
		return 12
	case (c == 0x02010                ) ||
		 (c >= 0x02013 && c <= 0x02016) ||
		 (c >= 0x02018 && c <= 0x02019) ||
		 (c >= 0x0201C && c <= 0x0201D) ||
		 (c >= 0x02025 && c <= 0x02026) ||
		 (c >= 0x02030 && c <= 0x02033) ||
		 (c == 0x02035                ) ||
		 (c == 0x0203B                ):
		return 24
	case (c >= 0x01100 && c <= 0x011FF) ||
		 (c >= 0x02460 && c <= 0x024FF) ||
		 (c >= 0x025A0 && c <= 0x027BF) ||
		 (c >= 0x02E80 && c <= 0x02FDF) ||
		 (c >= 0x03000 && c <= 0x0318F) ||
		 (c >= 0x031A0 && c <= 0x031EF) ||
		 (c >= 0x03200 && c <= 0x09FFF) ||
		 (c >= 0x0AC00 && c <= 0x0D7FF) ||
		 (c >= 0x0F900 && c <= 0x0FAFF) ||
		 (c >= 0x0FE30 && c <= 0x0FE4F) ||
		 (c >= 0x1F000 && c <= 0x1F9FF):
		return 24
	case (c >= 0x0FF01 && c <= 0x0FF5E) ||
		 (c >= 0x0FFE0 && c <= 0x0FFE5):
		return 24
	default:
		return 0
	}
}

func WidthOfString(str string) int {
	width := 0
	for _, c := range str {
		width += WidthOfChar(c)
	}
	return width * charHSize
}

func ColumnWidthWithAlignment(width, alignment int) int {
	return (width & 0xFFFF) | ((alignment & 3) << 16)
}

func ColumnWidth(v int) int {
	return v & 0xFFFF
}

func ColumnAlignment(v int) int {
	return (v >> 16) & 3
}

/*
 * 设置每列的宽度和对齐方式。
 */
func SetColumnWidths(widths ...int) {
	for i := 0; i < MAX_COLUMNS; i++ {
		widthOfColumns[i] = 0
	}

	remain := DOTS_PER_LINE // 未分配的宽度
	for i, v := range widths {
		if i == MAX_COLUMNS { // 超过最大支持的列数
			return
		}
		width := ColumnWidth(v)
		alignment := ColumnAlignment(v)
		if width == 0 || width > remain { // 使用所有未分配的宽度作为最后一列
			widthOfColumns[i] = ColumnWidthWithAlignment(remain, alignment)
			return
		}
		widthOfColumns[i] = v
		remain -= width
	}
}

/*
 * 设置每列的标志位。
 */
func SetColumnFlags(flags ...int) {
	for i := 0; i < MAX_COLUMNS; i++ {
		flagOfColumns[i] = 0
	}

	for i, v := range flags {
		if i == MAX_COLUMNS { // 超过最大支持的列数
			return
		}
		flagOfColumns[i] = v
	}
}

/*
 * 按当前的列设置进行分列打印。
 */
func PrintInColumns(texts ...string) {
	var substrs [MAX_COLUMNS][]string
	var substrwidths [MAX_COLUMNS][]int

	for i, str := range texts {
		if i == MAX_COLUMNS || widthOfColumns[i] == 0 {
			break
		}
		substrs[i] = append(substrs[i], "")
		substrwidths[i] = append(substrwidths[i], 0)
		row := 0
		for _, c := range str {
			if c == '\n' {
				substrs[i] = append(substrs[i], "")
				substrwidths[i] = append(substrwidths[i], 0)
				row++
			} else {
				w := WidthOfChar(c)
				if w == 0 {
					c = '?'
					w = 12
				}
				w *= charHSize
				if substrwidths[i][row] + w > ColumnWidth(widthOfColumns[i]) {
					substrs[i] = append(substrs[i], "")
					substrwidths[i] = append(substrwidths[i], 0)
					row++
				}
				substrs[i][row] += string(c)
				substrwidths[i][row] += w
			}
		}
	}

	maxrows := 0
	for i := 0; i < MAX_COLUMNS && widthOfColumns[i] != 0; i++ {
		if maxrows < len(substrs[i]) {
			maxrows = len(substrs[i])
		}
	}

	for row := 0; row < maxrows; row++ {
		pos := 0
		for i := 0; i < MAX_COLUMNS && widthOfColumns[i] != 0; i++ {
			width := ColumnWidth(widthOfColumns[i])
			alignment := ColumnAlignment(widthOfColumns[i])
			if row < len(substrs[i]) {
				switch alignment {
				case 1:
					SetAbsolutePrintPosition(pos + (width - substrwidths[i][row]) / 2)
				case 2:
					SetAbsolutePrintPosition(pos + (width - substrwidths[i][row]))
				default:
				}
				if (flagOfColumns[i] & COLUMN_FLAG_BW_REVERSE) != 0 {
					SetBlackWhiteReverseMode(true)
				}
				orderContent += hex.EncodeToString([]byte(substrs[i][row]))
				if (flagOfColumns[i] & COLUMN_FLAG_BW_REVERSE) != 0 {
					SetBlackWhiteReverseMode(false)
				}
			}
			pos += width
			SetAbsolutePrintPosition(pos)
		}
		orderContent += "0a"
	}
}

//////////////////////////////////////////////////
// 条码/二维码打印
//////////////////////////////////////////////////

/*
 * 在小票内容中追加条码。
 */
func AppendBarcode(hri_pos, height, module_size, barcode_type int, text string) {
	text_length := len(text)
	if text_length == 0 {
		return
	}
	if text_length > 255 {
		text_length = 255
	}
	if height < 1 {
		height = 1
	} else if height > 255 {
		height = 255
	}
	if module_size < 1 {
		module_size = 1
	} else if module_size > 6 {
		module_size = 6
	}

	orderContent += fmt.Sprintf("1d48%02x", (hri_pos & 3))
	orderContent += "1d6600"
	orderContent += fmt.Sprintf("1d68%02x", height)
	orderContent += fmt.Sprintf("1d77%02x", module_size)
	orderContent += fmt.Sprintf("1d6b%02x%02x", barcode_type, text_length)

	content := []byte(text)
	for i := 0; i < text_length; i++ {
		orderContent += fmt.Sprintf("%02x", content[i])
	}
}

/*
 * 在小票内容中追加二维码。
 */
func AppendQRcode(module_size, ec_level int, text string) {
	text_length := len(text)
	if text_length == 0 {
		return
	}
	if text_length > 65535 {
		text_length = 65535
	}
	if module_size < 1 {
		module_size = 1
	} else if module_size > 16 {
		module_size = 16
	}
	if ec_level < 0 {
		ec_level = 0
	} else if ec_level > 3 {
		ec_level = 3
	}

	orderContent += "1d286b040031410000"
	orderContent += fmt.Sprintf("1d286b03003143%02x", module_size)
	orderContent += fmt.Sprintf("1d286b03003145%02x", ec_level + 48)
	orderContent += fmt.Sprintf("1d286b%02x%02x315030", ((text_length + 3) & 0xFF), (((text_length + 3) >> 8) & 0xFF))

	content := []byte(text)
	for i := 0; i < text_length; i++ {
		orderContent += fmt.Sprintf("%02x", content[i])
	}

	orderContent += "1d286b0300315130"
}

//////////////////////////////////////////////////
// 图像打印
//////////////////////////////////////////////////

/*
 * 灰阶转单色 - 弥漫仿色算法。
 */
func DiffuseDither(src_data []byte, width, height int) []byte {
	if width <= 0 || height <= 0 {
		return nil
	}
	if len(src_data) < width * height {
		return nil
	}

	bmwidth := (width + 7) / 8
	dst_data := make([]byte, bmwidth * height)

	var line_buffer [2][]int
	var i, p, q, x, y int
	var line1, line2, b1, b2 int
	var err, e1, e3, e5, e7 int
	var mask byte
	var not_last_line bool

	line_buffer[0] = make([]int, width)
	line_buffer[1] = make([]int, width)

	line1 = 0
	line2 = 1

	for i = 0; i < width; i++ {
		line_buffer[0][i] = 0
		line_buffer[1][i] = int(src_data[i])
	}

	for y = 0; y < height; y++ {
		line1, line2 = line2, line1
		if y < height - 1 {
			not_last_line = true
		} else {
			not_last_line = false
		}

		if not_last_line {
			p = (y + 1) * width
			for i = 0; i < width; i++ {
				line_buffer[line2][i] = int(src_data[p])
				p++
			}
		}

		q = y * bmwidth
		for i = 0; i < bmwidth; i++ {
			dst_data[q] = 0
			q++
		}

		b1 = 0
		b2 = 0
		q = y * bmwidth
		mask = 0x80

		for x = 1; x <= width; x++ {
			if line_buffer[line1][b1] < 128 { // Black pixel
				err = line_buffer[line1][b1]
				dst_data[q] |= mask
			} else {
				err = line_buffer[line1][b1] - 255
			}
			b1++
			if mask == 1 {
				q++
				mask = 0x80
			} else {
				mask >>= 1
			}
			e7 = ((err * 7) + 8) >> 4
			e5 = ((err * 5) + 8) >> 4
			e3 = ((err * 3) + 8) >> 4
			e1 = err - (e7 + e5 + e3)
			if x < width {
				line_buffer[line1][b1] += e7
			}
			if not_last_line {
				line_buffer[line2][b2] += e5
				if x > 1 {
					line_buffer[line2][b2 - 1] += e3
				}
				if x < width {
					line_buffer[line2][b2 + 1] += e1
				}
			}
			b2++
		}
	}

	return dst_data
}

/*
 * 灰阶转单色 - 门限值仿色算法。
 */
func ThresholdDither(src_data []byte, width, height int) []byte {
	if width <= 0 || height <= 0 {
		return nil
	}
	if len(src_data) < width * height {
		return nil
	}

	bmwidth := (width + 7) / 8
	dst_data := make([]byte, bmwidth * height)

	var p, q, k, x, y int
	var mask byte

	p = 0
	q = 0

	for y = 0; y < height; y++ {
		k = q
		mask = 0x80
		for x = 0; x < width; x++ {
			if src_data[p] < 128 { // Black pixel
				dst_data[k] |= mask
			}
			p++
			if mask == 1 {
				k++
				mask = 0x80
			} else {
				mask >>= 1
			}
		}
		q += bmwidth
	}

	return dst_data
}

/*
 * RGB转灰阶。
 */
func ConvertToGray(img image.Image, width, height int) []byte {
	gray_data := make([]byte, width * height)
	i := 0
	for y := 0; y < height; y++ {
		for x := 0; x < width; x++ {
			r, g, b, _ := img.At(x, y).RGBA()
			r = (r >> 8) & 0xFF
			g = (g >> 8) & 0xFF
			b = (b >> 8) & 0xFF
			gray_data[i] = byte((r * 11 + g * 16 + b * 5) / 32)
			i++
		}
	}
	return gray_data
}

/*
 * 图像等比缩小。
 */
func ResizeImage(img image.Image) (image.Image, int, int) {
	bounds := img.Bounds()
	w := bounds.Max.X
	h := bounds.Max.Y
	if w > DOTS_PER_LINE {
		h = DOTS_PER_LINE * h / w
		w = DOTS_PER_LINE
		return resize.Resize(uint(w), uint(h), img, resize.Bicubic), w, h
	} else {
		return img, w, h
	}
}

/*
 * 在小票内容中追加图像。
 */
func AppendImage(imgfile string) {
	/* 打开图像文件，可以为 jpg/png/gif 格式 */
	f, err := os.Open(imgfile)
	if err != nil {
		fmt.Printf("os.Open: %s\n", err.Error())
		return
	}
	defer f.Close()

	img, _, err := image.Decode(f)
	if err != nil {
		fmt.Printf("image.Decode: %s\n", err.Error())
		return
	}

	/* 如果图像宽度大于热敏纸宽度，进行等比缩小 */
	img, w, h := ResizeImage(img)

	/* 将彩色图像转换为 8bpp 灰阶数据 */
	gray_data := ConvertToGray(img, w, h)

	/* 将 8bpp 灰阶数据转换为 1bpp 单色数据。
	 * 可以用 DiffuseDither 或 ThresholdDither 算法。
	 * 对于色彩比较单一的图案，用 ThresholdDither 效果更好 */
	mono_data := DiffuseDither(gray_data, w, h)

	/* 构造 "GS v 0" 图像打印指令 */
	w = (w + 7) / 8
	orderContent += "1d763000"
	orderContent += fmt.Sprintf("%02x%02x", (w & 0xFF), ((w >> 8) & 0xFF))
	orderContent += fmt.Sprintf("%02x%02x", (h & 0xFF), ((h >> 8) & 0xFF))
	orderContent += hex.EncodeToString(mono_data)
}

//////////////////////////////////////////////////
// 页模式指令
//////////////////////////////////////////////////

/*
 * 进入页模式。
 */
func EnterPageMode() {
	orderContent += "1b4c"
}

/*
 * 设置页模式打印区域。
 * x, y: 打印区域原点（区域的左上角）坐标
 * w, h: 打印区域的宽度和高度
 */
func SetPrintAreaInPageMode(x, y, w, h int) {
	orderContent += "1b57"
	orderContent += fmt.Sprintf("%02x%02x", (x & 0xFF), ((x >> 8) & 0xFF))
	orderContent += fmt.Sprintf("%02x%02x", (y & 0xFF), ((y >> 8) & 0xFF))
	orderContent += fmt.Sprintf("%02x%02x", (w & 0xFF), ((w >> 8) & 0xFF))
	orderContent += fmt.Sprintf("%02x%02x", (h & 0xFF), ((h >> 8) & 0xFF))
}

/*
 * 设置页模式打印方向。
 * dir: 0:画布不旋转; 1:画布顺时针旋转90度; 2:画布顺时针旋转180度; 3:画布顺时针旋转270度
 */
func SetPrintDirectionInPageMode(dir int) {
	if dir >= 0 && dir <= 3 {
		orderContent += fmt.Sprintf("1b54%02x", dir)
	}
}

/*
 * 设置页模式下的绝对垂直打印位置。
 */
func SetAbsolutePrintPositionInPageMode(n int) {
	if n >= 0 && n <= 65535 {
		orderContent += fmt.Sprintf("1d24%02x%02x", (n & 0xFF), ((n >> 8) & 0xFF))
	}
}

/*
 * 设置页模式下的相对垂直打印位置。
 */
func SetRelativePrintPositionInPageMode(n int) {
	if n >= -32768 && n <= 32767 {
		orderContent += fmt.Sprintf("1d5c%02x%02x", (n & 0xFF), ((n >> 8) & 0xFF))
	}
}

/*
 * 打印并退出页模式。
 */
func PrintAndExitPageMode() {
	orderContent += "0c"
}

/*
 * 打印（维持在页模式）。
 */
func PrintInPageMode() {
	orderContent += "1b0c"
}

/*
 * 清除打印区域（维持在页模式）。
 */
func ClearInPageMode() {
	orderContent += "18"
}

/*
 * 不打印，退出页模式。
 */
func ExitPageMode() {
	orderContent += "1b53"
}

func main() {
	LineFeed(1)

	SetLineSpacing(80)
	SetPrintModes(true, true, false)
	SetAlignment(ALIGN_CENTER)
	AppendText("*** 打印测试 ***\n")

	RestoreDefaultLineSpacing()
	SetPrintModes(false, false, false)
	SetAlignment(ALIGN_LEFT)

	/* 设置三列。第一列96点，左对齐；第二列144点，居中对齐；第三列使用剩余的宽度，右对齐 */
	SetColumnWidths(
		ColumnWidthWithAlignment(96,  ALIGN_LEFT  ),
		ColumnWidthWithAlignment(144, ALIGN_CENTER),
		ColumnWidthWithAlignment(0,   ALIGN_RIGHT ))
	SetColumnFlags(
		0,
		0,
		COLUMN_FLAG_BW_REVERSE)

	PrintInColumns("商品名称", "数量\n(单位：随意)", "小计\n(单位：元)")
	LineFeed(1)
	PrintInColumns("这是\"一个很长的品名\"", "x1000", "￥2020.99")
	LineFeed(1)
	PrintInColumns("橙子", "【备注：赠品购物满1000,000元送一只】", "￥0.00")
	LineFeed(1)

	SetAlignment(ALIGN_CENTER)

	/* 打印 CODE128 条码 */
	AppendBarcode(HRI_POS_BELOW, 160, 3, 73, "Abc-000789")
	LineFeed(1)

	/* 打印二维码 */
	AppendQRcode(5, 1, "https://docs.sunmi.com/docking-debugging/商米云打印机合作伙伴对接说明/1-了解一下对接流程/")
	LineFeed(1)

	SetAlignment(ALIGN_LEFT)

	/* 打印图像文件 */
	if len(os.Args) > 1 {
		AppendImage(os.Args[1])
	}
	LineFeed(1)

	/* 页模式打印 */
	SetAlignment(ALIGN_CENTER)
	AppendText("---- 页模式多区域打印 ----\n")
	SetAlignment(ALIGN_LEFT)
	EnterPageMode()
	// 区域1
	SetPrintAreaInPageMode(0, 0, 144, 500)
	SetPrintDirectionInPageMode(0)
	AppendText("永和九年，岁在癸丑，暮春之初，会于会稽山阴之兰亭，修禊事也。群贤毕至，少长咸集。" +
		"此地有崇山峻岭，茂林修竹；又有清流激湍，映带左右，引以为流觞曲水，列坐其次。\n")
	// 区域2
	SetPrintAreaInPageMode(156, 0, 144, 500)
	SetPrintDirectionInPageMode(2)
	AppendText("鎌倉アナウンサーはまず流暢な中国語でアナウンサーとしての豊富な経験を紹介されました。\n")
	// 区域3
	SetPrintAreaInPageMode(312, 0, 72, 500)
	SetPrintDirectionInPageMode(3)
	AppendText("Scarlett is a woman who can deal with a nation at war, Atlanta burning.\n")
	// 打印并退出页模式
	PrintAndExitPageMode()

	LineFeed(4)
	CutPaper(false)

	PushToPrinter("N403D06A00023")
}
